@@ -46,6 +46,7 @@ type TheTlsf<Options> =
     FlexTlsf<os::Source<Options>, usize, usize, { USIZE_BITS as usize }, { USIZE_BITS as usize }>;
 
 impl<Options: GlobalTlsfOptions> Init for GlobalTlsf<Options> {
+    #[allow(clippy::clippy::declare_interior_mutable_const)]
     const INIT: Self = Self::INIT;
 }
 
@@ -94,6 +95,7 @@ unsafe impl<Options: GlobalTlsfOptions> Sync for GlobalTlsf<Options> {}
 
 impl<Options: GlobalTlsfOptions> GlobalTlsf<Options> {
     /// The initializer.
+    #[allow(clippy::clippy::declare_interior_mutable_const)]
     pub const INIT: Self = Self {
         inner: UnsafeCell::new(Init::INIT),
         mutex: Init::INIT,
@@ -171,23 +173,92 @@ unsafe impl<Options: GlobalTlsfOptions> alloc::GlobalAlloc for GlobalTlsf<Option
                 .reallocate(ptr, new_layout)
                 .map(NonNull::as_ptr)
                 .unwrap_or(ptr::null_mut())
+        } else if let Some(new_ptr) = inner.allocate(new_layout) {
+            // Safety: the previously allocated block cannot overlap the
+            //         newly allocated block.
+            //         The safety contract for `deallocate` must be upheld
+            //         by the caller.
+            ptr::copy_nonoverlapping(ptr.as_ptr(), new_ptr.as_ptr(), layout.size().min(new_size));
+            inner.deallocate(ptr, layout.align());
+            new_ptr.as_ptr()
         } else {
-            // Safety: the caller must ensure that `new_layout` is greater than zero.
-            if let Some(new_ptr) = inner.allocate(new_layout) {
-                // Safety: the previously allocated block cannot overlap the
-                //         newly allocated block.
-                //         The safety contract for `deallocate` must be upheld
-                //         by the caller.
-                ptr::copy_nonoverlapping(
-                    ptr.as_ptr(),
-                    new_ptr.as_ptr(),
-                    layout.size().min(new_size),
-                );
-                inner.deallocate(ptr, layout.align());
-                new_ptr.as_ptr()
-            } else {
-                ptr::null_mut()
-            }
+            ptr::null_mut()
         }
     }
 }
+
+/// Provides allocation functions modelled after the standard C and POSIX
+/// allocation functions (e.g., `malloc`, `memalign`).
+///
+/// Note that this trait might require a less efficient implementation than
+/// [`core::alloc::GlobalAlloc`]. This applies to [`GlobalTlsf`].
+pub unsafe trait CAlloc {
+    /// Allocate a memory block.
+    ///
+    /// Returns the starting address of the allocated memory block on success;
+    /// `None` otherwise.
+    fn allocate(&self, layout: alloc::Layout) -> Option<NonNull<u8>>;
+
+    /// Deallocate a previously allocated memory block.
+    ///
+    /// # Safety
+    ///
+    ///  - `ptr` must denote a memory block previously allocated by calling
+    ///    [`Self::allocate`] on `self`.
+    ///
+    unsafe fn deallocate(&self, ptr: NonNull<u8>);
+
+    /// Shrink or grow a previously allocated memory block.
+    ///
+    /// Returns the new starting address of the memory block on success;
+    /// `None` otherwise.
+    ///
+    /// # Safety
+    ///
+    ///  - `ptr` must denote a memory block previously allocated by calling
+    ///    [`Self::allocate`] on `self`.
+    ///
+    unsafe fn reallocate(&self, ptr: NonNull<u8>, new_layout: alloc::Layout)
+        -> Option<NonNull<u8>>;
+}
+
+unsafe impl<Options: GlobalTlsfOptions> CAlloc for GlobalTlsf<Options> {
+    fn allocate(&self, layout: alloc::Layout) -> Option<NonNull<u8>> {
+        let mut inner = self.lock_inner();
+        inner.allocate(layout)
+    }
+
+    unsafe fn deallocate(&self, ptr: NonNull<u8>) {
+        let mut inner = self.lock_inner();
+        // Safety: `ptr` denotes a previous allocation
+        inner.deallocate_unknown_align(ptr);
+    }
+
+    unsafe fn reallocate(
+        &self,
+        ptr: NonNull<u8>,
+        new_layout: alloc::Layout,
+    ) -> Option<NonNull<u8>> {
+        let mut inner = self.lock_inner();
+        if let Some(new_ptr) = inner.allocate(new_layout) {
+            // Safety: `ptr` denotes a previous allocation
+            let old_size = TheTlsf::<Options>::size_of_allocation_unknown_align(ptr);
+            // Safety: the previously allocated block cannot overlap the
+            //         newly allocated block.
+            //         The safety contract for `deallocate` must be upheld
+            //         by the caller.
+            ptr::copy_nonoverlapping(
+                ptr.as_ptr(),
+                new_ptr.as_ptr(),
+                new_layout.size().min(old_size),
+            );
+            inner.deallocate_unknown_align(ptr);
+            Some(new_ptr)
+        } else {
+            None
+        }
+    }
+}
+
+#[cfg(test)]
+mod tests;
