@@ -6,7 +6,7 @@ use core::{
     marker::PhantomData,
     mem::{self, MaybeUninit},
     num::NonZeroUsize,
-    ptr::NonNull,
+    ptr::{addr_of, NonNull},
 };
 
 use crate::{
@@ -100,6 +100,7 @@ pub(crate) const USIZE_BITS: u32 = core::mem::size_of::<usize>() as u32 * 8;
 // The header is actually aligned at `size_of::<usize>() * 4`-byte boundaries
 // but the alignment is set to a half value here not to introduce a padding at
 // the end of this struct.
+#[repr(C)]
 #[cfg_attr(target_pointer_width = "16", repr(align(4)))]
 #[cfg_attr(target_pointer_width = "32", repr(align(8)))]
 #[cfg_attr(target_pointer_width = "64", repr(align(16)))]
@@ -166,14 +167,29 @@ struct FreeBlockHdr {
 /// The payload immediately follows this header. However, if the alignment
 /// requirement is greater than or equal to [`GRANULARITY`], an up to
 /// `align - GRANULARITY / 2` bytes long padding will be inserted between them,
-/// and the last word of the padding will encode where the header is located.
-/// (See `used_block_hdr_for_allocation`)
+/// and the last part of the padding ([`UsedBlockPad`]) will encode where the
+/// header is located.
 #[repr(C)]
 #[derive(Debug)]
 struct UsedBlockHdr {
     common: BlockHdr,
 }
 
+/// In a used memory block with an alignment requirement larger than or equal to
+/// `GRANULARITY`, the payload is preceded by this structure.
+#[derive(Debug)]
+#[repr(C)]
+struct UsedBlockPad {
+    block_hdr: NonNull<UsedBlockHdr>,
+}
+
+impl UsedBlockPad {
+    #[inline]
+    fn get_for_allocation(ptr: NonNull<u8>) -> *mut Self {
+        ptr.cast::<Self>().as_ptr().wrapping_sub(1)
+    }
+}
+
 impl<FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, const SLLEN: usize> Default
     for Tlsf<'_, FLBitmap, SLBitmap, FLLEN, SLLEN>
 {
@@ -245,16 +261,13 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
         debug_assert!(size % GRANULARITY == 0);
         let fl = USIZE_BITS - GRANULARITY_LOG2 - 1 - size.leading_zeros();
 
-        let sl = if GRANULARITY_LOG2 < Self::SLI && fl < Self::SLI - GRANULARITY_LOG2 {
-            size << ((Self::SLI - GRANULARITY_LOG2) - fl)
-        } else {
-            let sl = size >> (fl + GRANULARITY_LOG2 - Self::SLI);
+        // The shift amount can be negative, and rotation lets us handle both
+        // cases without branching. Underflowed digits can be simply masked out
+        // in `map_floor`.
+        let sl = size.rotate_right((fl + GRANULARITY_LOG2).wrapping_sub(Self::SLI));
 
-            // The most significant one of `size` should be at `sl[SLI]`
-            debug_assert!((sl >> Self::SLI) == 1);
-
-            sl
-        };
+        // The most significant one of `size` should be now at `sl[SLI]`
+        debug_assert!(((sl >> Self::SLI) & 1) == 1);
 
         // `fl` must be in a valid range
         if fl as usize >= FLLEN {
@@ -272,21 +285,19 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
         debug_assert!(size % GRANULARITY == 0);
         let mut fl = USIZE_BITS - GRANULARITY_LOG2 - 1 - size.leading_zeros();
 
-        let sl = if GRANULARITY_LOG2 < Self::SLI && fl < Self::SLI - GRANULARITY_LOG2 {
-            size << ((Self::SLI - GRANULARITY_LOG2) - fl)
-        } else {
-            let mut sl = size >> (fl + GRANULARITY_LOG2 - Self::SLI);
+        // The shift amount can be negative, and rotation lets us handle both
+        // cases without branching.
+        let mut sl = size.rotate_right((fl + GRANULARITY_LOG2).wrapping_sub(Self::SLI));
 
-            // round up (this is specific to `map_ceil`)
-            sl += (sl << (fl + GRANULARITY_LOG2 - Self::SLI) != size) as usize;
+        // The most significant one of `size` should be now at `sl[SLI]`
+        debug_assert!(((sl >> Self::SLI) & 1) == 1);
 
-            debug_assert!((sl >> Self::SLI) == 0b01 || (sl >> Self::SLI) == 0b10);
+        // Underflowed digits appear in `sl[SLI + 1..USIZE-BITS]`. They should
+        // be rounded up
+        sl = (sl & (SLLEN - 1)) + (sl >= (1 << (Self::SLI + 1))) as usize;
 
-            // if sl[SLI + 1] { fl += 1; sl = 0; }
-            fl += (sl >> (Self::SLI + 1)) as u32;
-
-            sl
-        };
+        // if sl[SLI] { fl += 1; sl = 0; }
+        fl += (sl >> Self::SLI) as u32;
 
         // `fl` must be in a valid range
         if fl as usize >= FLLEN {
@@ -770,7 +781,7 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
             let (fl, sl) = self.search_suitable_free_block_list_for_allocation(search_size)?;
 
             // Get a free block: `block`
-            let first_free = &mut self.first_free[fl][sl];
+            let first_free = self.first_free.get_unchecked_mut(fl).get_unchecked_mut(sl);
             let block = first_free.unwrap_or_else(|| {
                 debug_assert!(false, "bitmap outdated");
                 // Safety: It's unreachable
@@ -790,8 +801,9 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
                 next_free.as_mut().prev_free = None;
             } else {
                 // The free list is now empty - update the bitmap
-                self.sl_bitmap[fl].clear_bit(sl as u32);
-                if self.sl_bitmap[fl] == SLBitmap::ZERO {
+                let sl_bitmap = self.sl_bitmap.get_unchecked_mut(fl);
+                sl_bitmap.clear_bit(sl as u32);
+                if *sl_bitmap == SLBitmap::ZERO {
                     self.fl_bitmap.clear_bit(fl as u32);
                 }
             }
@@ -808,11 +820,6 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
                 debug_assert_ne!(unaligned_ptr, ptr.as_ptr() as usize);
             }
 
-            // Place a header pointer (used by `used_block_hdr_for_allocation`)
-            if layout.align() >= GRANULARITY {
-                *ptr.cast::<NonNull<_>>().as_ptr().sub(1) = block;
-            }
-
             // Calculate the actual overhead and the final block size of the
             // used block being created here
             let overhead = ptr.as_ptr() as usize - block.as_ptr() as usize;
@@ -853,6 +860,11 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
             let mut block = block.cast::<UsedBlockHdr>();
             block.as_mut().common.size = new_size | SIZE_USED;
 
+            // Place a `UsedBlockPad` (used by `used_block_hdr_for_allocation`)
+            if layout.align() >= GRANULARITY {
+                (*UsedBlockPad::get_for_allocation(ptr)).block_hdr = block;
+            }
+
             Some(ptr)
         }
     }
@@ -908,13 +920,66 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
         align: usize,
     ) -> NonNull<UsedBlockHdr> {
         if align >= GRANULARITY {
-            // Read the header pointer at `ptr - USIZE_LEN`
-            *ptr.cast::<NonNull<UsedBlockHdr>>().as_ptr().sub(1)
+            // Read the header pointer
+            (*UsedBlockPad::get_for_allocation(ptr)).block_hdr
         } else {
             NonNull::new_unchecked(ptr.as_ptr().sub(GRANULARITY / 2)).cast()
         }
     }
 
+    /// Find the `UsedBlockHdr` for an allocation (any `NonNull<u8>` returned by
+    /// our allocation functions) with an unknown alignment.
+    ///
+    /// Unlike `used_block_hdr_for_allocation`, this function does not require
+    /// knowing the allocation's alignment but might be less efficient.
+    ///
+    /// # Safety
+    ///
+    ///  - `ptr` must point to an allocated memory block returned by
+    ///      `Self::{allocate, reallocate}`.
+    ///
+    #[inline]
+    unsafe fn used_block_hdr_for_allocation_unknown_align(
+        ptr: NonNull<u8>,
+    ) -> NonNull<UsedBlockHdr> {
+        // Case 1: `align >= GRANULARITY`
+        let c1_block_hdr_ptr: *const NonNull<UsedBlockHdr> =
+            addr_of!((*UsedBlockPad::get_for_allocation(ptr)).block_hdr);
+        // Case 2: `align < GRANULARITY`
+        let c2_block_hdr = ptr.cast::<UsedBlockHdr>().as_ptr().wrapping_sub(1);
+        let c2_prev_phys_block_ptr: *const Option<NonNull<BlockHdr>> =
+            addr_of!((*c2_block_hdr).common.prev_phys_block);
+
+        // They are both present at the same location, so we can be assured that
+        // their contents are initialized and we can read them safely without
+        // knowing which case applies first.
+        debug_assert_eq!(
+            c1_block_hdr_ptr as *const usize,
+            c2_prev_phys_block_ptr as *const usize
+        );
+
+        // Read it as `Option<NonNull<BlockHdr>>`.
+        if let Some(block_ptr) = *c2_prev_phys_block_ptr {
+            // Where does the block represented by `block_ptr` end?
+            // (Note: `block_ptr.size` might include `SIZE_USED`.)
+            let block_end = block_ptr.as_ptr() as usize + block_ptr.as_ref().size;
+
+            if ptr.as_ptr() as usize > block_end {
+                // The block represented by `block_ptr` does not include `ptr`.
+                // It's Case 2.
+                NonNull::new_unchecked(c2_block_hdr)
+            } else {
+                // `ptr` is inside the block - it's Case 1.
+                // (Note: `ptr == block_end` should count as being inside
+                // because the payload might be zero-sized.)
+                *c1_block_hdr_ptr
+            }
+        } else {
+            // It's non-nullable in Case 1, so we can rule out Case 1.
+            NonNull::new_unchecked(c2_block_hdr)
+        }
+    }
+
     /// Deallocate a previously allocated memory block.
     ///
     /// # Time Complexity
@@ -930,7 +995,34 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
     pub unsafe fn deallocate(&mut self, ptr: NonNull<u8>, align: usize) {
         // Safety: `ptr` is a previously allocated memory block with the same
         //         alignment as `align`. This is upheld by the caller.
-        let mut block = Self::used_block_hdr_for_allocation(ptr, align).cast::<BlockHdr>();
+        let block = Self::used_block_hdr_for_allocation(ptr, align).cast::<BlockHdr>();
+        self.deallocate_block(block);
+    }
+
+    /// Deallocate a previously allocated memory block with an unknown alignment.
+    ///
+    /// Unlike `deallocate`, this function does not require knowing the
+    /// allocation's alignment but might be less efficient.
+    ///
+    /// # Time Complexity
+    ///
+    /// This method will complete in constant time.
+    ///
+    /// # Safety
+    ///
+    ///  - `ptr` must denote a memory block previously allocated via `self`.
+    ///
+    pub(crate) unsafe fn deallocate_unknown_align(&mut self, ptr: NonNull<u8>) {
+        // Safety: `ptr` is a previously allocated memory block. This is upheld
+        //         by the caller.
+        let block = Self::used_block_hdr_for_allocation_unknown_align(ptr).cast::<BlockHdr>();
+        self.deallocate_block(block);
+    }
+
+    /// Deallocate a previously allocated memory block. Takes a pointer to
+    /// `BlockHdr` instead of a payload pointer.
+    #[inline]
+    unsafe fn deallocate_block(&mut self, mut block: NonNull<BlockHdr>) {
         let mut size = block.as_ref().size & !SIZE_USED;
         debug_assert!((block.as_ref().size & SIZE_USED) != 0);
 
@@ -1006,7 +1098,7 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
     ///
     /// # Safety
     ///
-    ///  - `ptr` must denote a memory block previously allocated via `self`.
+    ///  - `ptr` must denote a memory block previously allocated via `Self`.
     ///  - The memory block must have been allocated with the same alignment
     ///    ([`Layout::align`]) as `align`.
     ///
@@ -1024,6 +1116,28 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
         block_end - payload_start
     }
 
+    /// Get the payload size of the allocation with an unknown alignment. The
+    /// returned size might be larger than the size specified at the allocation
+    /// time.
+    ///
+    /// # Safety
+    ///
+    ///  - `ptr` must denote a memory block previously allocated via `Self`.
+    ///
+    #[inline]
+    pub(crate) unsafe fn size_of_allocation_unknown_align(ptr: NonNull<u8>) -> usize {
+        // Safety: `ptr` is a previously allocated memory block.
+        //         This is upheld by the caller.
+        let block = Self::used_block_hdr_for_allocation_unknown_align(ptr);
+
+        let size = block.as_ref().common.size - SIZE_USED;
+        debug_assert_eq!(size, block.as_ref().common.size & SIZE_SIZE_MASK);
+
+        let block_end = block.as_ptr() as usize + size;
+        let payload_start = ptr.as_ptr() as usize;
+        block_end - payload_start
+    }
+
     // TODO: `reallocate_no_move` (constant-time reallocation)
 
     /// Shrink or grow a previously allocated memory block.
@@ -1327,7 +1441,7 @@ impl<'pool, FLBitmap: BinInteger, SLBitmap: BinInteger, const FLLEN: usize, cons
 
         // Place a header pointer (used by `used_block_hdr_for_allocation`)
         if new_layout.align() >= GRANULARITY {
-            *new_ptr.cast::<NonNull<_>>().as_ptr().sub(1) = new_block;
+            (*UsedBlockPad::get_for_allocation(new_ptr)).block_hdr = new_block;
         }
 
         Some(new_ptr)
